@@ -39,14 +39,16 @@ class StepRunner:
             if self.lr_scheduler is not None:
                 self.lr_scheduler.step()
             self.optimizer.zero_grad()
+            
         all_preds = self.accelerator.gather(preds)
         all_labels = self.accelerator.gather(labels)
         all_loss = self.accelerator.gather(loss).sum()
-            
+        
         #metrics
         step_metrics = {self.stage+"_"+name:metric_fn(all_preds, all_labels).item() 
                         for name,metric_fn in self.metrics_dict.items()}
-        
+        if self.optimizer is not None and self.stage=="train":
+            step_metrics['lr'] = self.optimizer.state_dict()['param_groups'][0]['lr']
         return all_loss.item(),step_metrics
 
 class EpochRunner:
@@ -79,20 +81,21 @@ class EpochRunner:
             if i!=len(dataloader)-1:
                 loop.set_postfix(**step_log)
             else:
+                epoch_metrics = step_metrics
+                epoch_metrics.update({self.stage+"_"+name:metric_fn.compute().item() 
+                                 for name,metric_fn in self.steprunner.metrics_dict.items()})
                 epoch_loss = total_loss/step
-                epoch_metrics = {self.stage+"_"+name:metric_fn.compute().item() 
-                                 for name,metric_fn in self.steprunner.metrics_dict.items()}
                 epoch_log = dict({self.stage+"_loss":epoch_loss},**epoch_metrics)
                 loop.set_postfix(**epoch_log)
                 for name,metric_fn in self.steprunner.metrics_dict.items():
                     metric_fn.reset()
         return epoch_log
 
+
 class KerasModel(torch.nn.Module):
     def __init__(self,net,loss_fn,metrics_dict=None,optimizer=None,lr_scheduler = None):
         super().__init__()
-        self.net,self.loss_fn = net, loss_fn
-        self.metrics_dict = torch.nn.ModuleDict(metrics_dict) 
+        self.net,self.loss_fn,self.metrics_dict = net, loss_fn, torch.nn.ModuleDict(metrics_dict) 
         self.optimizer = optimizer if optimizer is not None else torch.optim.Adam(
             self.net.parameters(), lr=1e-3)
         self.lr_scheduler = lr_scheduler
@@ -101,99 +104,108 @@ class KerasModel(torch.nn.Module):
         return self.net.forward(x)
 
     def fit(self, train_data, val_data=None, epochs=10,ckpt_path='checkpoint.pt',
-            patience=5, monitor="val_loss", mode="min", mixed_precision='no'):
+            patience=5, monitor="val_loss", mode="min", mixed_precision='no',callbacks = None):
         
-        accelerator = Accelerator(mixed_precision=mixed_precision)
-        device = str(accelerator.device)
+        self.accelerator = Accelerator(mixed_precision=mixed_precision)
+        device = str(self.accelerator.device)
         device_type = '🐌'  if 'cpu' in device else '⚡️'
-        accelerator.print(colorful("<<<<<< "+device_type +" "+ device +" is used >>>>>>"))
+        self.accelerator.print(colorful("<<<<<< "+device_type +" "+ device +" is used >>>>>>"))
     
-        net,optimizer,lr_scheduler= accelerator.prepare(
-            self.net,self.optimizer,self.lr_scheduler)
-        train_dataloader,val_dataloader = accelerator.prepare(train_data,val_data)
+        self.net,self.loss_fn,self.metrics_dict,self.optimizer,self.lr_scheduler= self.accelerator.prepare(
+            self.net,self.loss_fn,self.metrics_dict,self.optimizer,self.lr_scheduler)
         
-        loss_fn = self.loss_fn
-        if isinstance(loss_fn,torch.nn.Module):
-            loss_fn.to(accelerator.device)
-        metrics_dict = self.metrics_dict 
-        metrics_dict.to(accelerator.device)
+        train_dataloader,val_dataloader = self.accelerator.prepare(train_data,val_data)
         
-        history = {}
+        self.history = {}
+        self.callbacks = self.accelerator.prepare(callbacks) if callbacks is not None else []
+        
+        if self.accelerator.is_local_main_process:
+            for callback_obj in self.callbacks:
+                callback_obj.on_fit_start(model = self)
         
         for epoch in range(1, epochs+1):
 
             nowtime = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
-            accelerator.print("\n"+"=========="*8 + "%s"%nowtime)
-            accelerator.print("Epoch {0} / {1}".format(epoch, epochs)+"\n")
+            self.accelerator.print("\n"+"=========="*8 + "%s"%nowtime)
+            self.accelerator.print("Epoch {0} / {1}".format(epoch, epochs)+"\n")
 
             # 1，train -------------------------------------------------  
             train_step_runner = StepRunner(
-                    net = net,
-                    loss_fn = loss_fn,
-                    accelerator = accelerator,
+                    net = self.net,
+                    loss_fn = self.loss_fn,
+                    accelerator = self.accelerator,
                     stage="train",
-                    metrics_dict=deepcopy(metrics_dict),
-                    optimizer = optimizer,
-                    lr_scheduler = lr_scheduler
+                    metrics_dict=deepcopy(self.metrics_dict),
+                    optimizer = self.optimizer,
+                    lr_scheduler = self.lr_scheduler
             )
 
             train_epoch_runner = EpochRunner(train_step_runner)
-            train_metrics = train_epoch_runner(train_dataloader)
+            train_metrics = {'epoch':epoch}
+            train_metrics.update(train_epoch_runner(train_dataloader))
+            
             for name, metric in train_metrics.items():
-                history[name] = history.get(name, []) + [metric]
+                self.history[name] = self.history.get(name, []) + [metric]
+                
+            if self.accelerator.is_local_main_process:
+                for callback_obj in self.callbacks:
+                    callback_obj.on_train_epoch_end(model = self)
 
             # 2，validate -------------------------------------------------
             if val_dataloader:
                 val_step_runner = StepRunner(
-                    net = net,
-                    loss_fn = loss_fn,
-                    accelerator = accelerator,
+                    net = self.net,
+                    loss_fn = self.loss_fn,
+                    accelerator = self.accelerator,
                     stage="val",
-                    metrics_dict= deepcopy(metrics_dict)
+                    metrics_dict= deepcopy(self.metrics_dict)
                 )
                 val_epoch_runner = EpochRunner(val_step_runner)
                 with torch.no_grad():
                     val_metrics = val_epoch_runner(val_dataloader)
 
-                val_metrics["epoch"] = epoch
                 for name, metric in val_metrics.items():
-                    history[name] = history.get(name, []) + [metric]
+                    self.history[name] = self.history.get(name, []) + [metric]
+                
+                if self.accelerator.is_local_main_process:
+                    for callback_obj in self.callbacks:
+                        callback_obj.on_validation_epoch_end(model = self)
 
             # 3，early-stopping -------------------------------------------------
-            accelerator.wait_for_everyone()
-            arr_scores = history[monitor]
+            self.accelerator.wait_for_everyone()
+            arr_scores = self.history[monitor]
             best_score_idx = np.argmax(arr_scores) if mode=="max" else np.argmin(arr_scores)
 
             if best_score_idx==len(arr_scores)-1:
-                unwrapped_net = accelerator.unwrap_model(net)
-                accelerator.save(unwrapped_net.state_dict(),ckpt_path)
-                accelerator.print(colorful("<<<<<< reach best {0} : {1} >>>>>>".format(monitor,
+                unwrapped_net = self.accelerator.unwrap_model(self.net)
+                self.accelerator.save(unwrapped_net.state_dict(),ckpt_path)
+                self.accelerator.print(colorful("<<<<<< reach best {0} : {1} >>>>>>".format(monitor,
                      arr_scores[best_score_idx])))
 
             if len(arr_scores)-best_score_idx>patience:
-                accelerator.print(colorful("<<<<<< {} without improvement in {} epoch, early stopping >>>>>>".format(
-                    monitor,patience)))
+                self.accelerator.print(colorful(
+                    "<<<<<< {} without improvement in {} epoch, early stopping >>>>>>".format(monitor,patience)))
                 break 
                 
-        if accelerator.is_local_main_process:
+        if self.accelerator.is_local_main_process:
+        
+            for callback_obj in self.callbacks:
+                callback_obj.on_fit_end(model = self)
+                
+            self.net = self.accelerator.unwrap_model(self.net)
             self.net.load_state_dict(torch.load(ckpt_path))
-            dfhistory = pd.DataFrame(history)
-            accelerator.print(dfhistory)
+            dfhistory = pd.DataFrame(self.history)
+            self.accelerator.print(dfhistory)
             return dfhistory 
     
     @torch.no_grad()
     def evaluate(self, val_data):
         accelerator = Accelerator()
-        self.net = accelerator.prepare(self.net)
+        self.net,self.loss_fn,self.metrics_dict = accelerator.prepare(self.net,self.loss_fn,self.metrics_dict)
         val_data = accelerator.prepare(val_data)
-        if isinstance(self.loss_fn,torch.nn.Module):
-            self.loss_fn.to(accelerator.device)
-        self.metrics_dict.to(accelerator.device)
-        
         val_step_runner = StepRunner(net = self.net,stage="val",
                     loss_fn = self.loss_fn,metrics_dict=deepcopy(self.metrics_dict),
                     accelerator = accelerator)
         val_epoch_runner = EpochRunner(val_step_runner)
         val_metrics = val_epoch_runner(val_data)
-        return val_metrics
-
+        return val_metrics
\ No newline at end of file
